#define PI 3.14159265

float DistributionGGX(vec3 N, vec3 H, float roughness)
{
    float a      = roughness*roughness;
    float a2     = a*a;
    float NdotH  = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;

    float nom   = a2;
    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;

    return nom / denom;
}

float GeometrySchlickGGX(float NdotV, float roughness)
{
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;

    float nom   = NdotV;
    float denom = NdotV * (1.0 - k) + k;

    return nom / denom;
}

float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
    float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx2  = GeometrySchlickGGX(NdotV, roughness);
    float ggx1  = GeometrySchlickGGX(NdotL, roughness);

    return ggx1 * ggx2;
}

vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
    return F0 + (1.0 - F0) * pow5(1.0 - cosTheta);
}

vec3 BRDF_Reflect(vec3 N, vec3 V, vec3 L, vec3 C, vec3 albedo, vec3 RFC, float MT, float SM)
{
    vec3 H = normalize(V + L);
    float D = DistributionGGX(N, H, 1.0 - SM);

    float cosTheta = dot(N, V);

    vec3 F0 = mix(vec3(0.04), albedo, MT);
    vec3 F = fresnelSchlick(cosTheta, F0);

    float G = GeometrySmith(N, V, L, 1.0 - SM);
    //float G = 1.0;

    vec3 ks = F;
    vec3 kd = (1.0 - ks) * (1.0 - MT);

    // vec3 nominator    = D * G * F;
    // float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0) + 0.001;
    // vec3 specular = nominator / denominator;

    return clamp(kd * C / PI + (D * G * F) / (4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0) + 0.001) * RFC, 0.0, 5.0)
            + ((1.0 - SM) * C + SM * F * RFC) * 0.5;
    //return C * (1.0 - SM) + RFC * F * SM;
}