#version 460 core

out vec4 texBuffer3;

#define LENSFLARE

uniform int isEyeInWater;
uniform int frameCounter;
uniform ivec2 eyeBrightnessSmooth;
uniform float flag_day;
uniform float flag_night;
uniform float flag_sunset;
uniform float aspectRatio;
uniform float rainStrength;
uniform float near;
uniform float far;
uniform vec2 screenTexel;
uniform vec3 sunVec;
uniform vec3 moonVec;
uniform vec3 shadowVec;
uniform vec3 upVec;

uniform vec3 skyTopColor;
uniform vec3 skyAddColor;
uniform vec3 skySunColor;
uniform vec3 skyCloudLight;
uniform vec3 skyCloudAmbient;
uniform vec3 waterFogColor;
uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex6;
uniform sampler2DShadow shadowtex1;

in vec2 Texcoord;

flat in float volumeRayStrength;
flat in vec3 worldSunVec;
flat in vec3 worldLightVec;

#ifdef LENSFLARE
flat in float sunVisibility;
flat in vec2 lf1Pos;
flat in vec2 lf2Pos;
flat in vec2 lf3Pos;
flat in vec2 lf4Pos;
#endif

#include "/HmwsLib/BasicFunctions.glsl"
#include "/HmwsLib/PackDatas.glsl"
#include "/HmwsLib/SpaceTransform.glsl"
#include "/HmwsLib/VolumeClouds.glsl"
#include "/HmwsLib/SkyBaseColor.glsl"
#include "/HmwsLib/RayTrac.glsl"
#include "/HmwsLib/VolumeRays.glsl"
#include "/HmwsLib/Dithering.glsl"
#include "/HmwsLib/BRDF.glsl"

#ifdef LENSFLARE

#define RAY_TRAC_REFRACT

#define MANHATTAN_DISTANCE(DELTA) abs(DELTA.x)+abs(DELTA.y)
 
#define LENS_FLARE(COLOR, UV, LFPOS, LFSIZE, LFCOLOR) { \
                vec2 delta = UV - LFPOS; delta.x *= aspectRatio; \
                if(MANHATTAN_DISTANCE(delta) < LFSIZE * 2.0) { \
                    float d = max(LFSIZE - sqrt(dot(delta, delta)), 0.0); \
                    COLOR += LFCOLOR.rgb * LFCOLOR.a * smoothstep(0.0, LFSIZE, d) * sunVisibility;\
                } }
 
#define LF1SIZE 0.1
#define LF2SIZE 0.15
#define LF3SIZE 0.25
#define LF4SIZE 0.25

vec3 lensFlare(vec3 color, vec2 uv)
{
    if(sunVisibility <= 0.0)
        return color;
    bool sunvvv = dot(sunVec, upVec) > 0.0;
        
    vec4 LF1COLOR = sunvvv ? vec4(1.0, 1.0, 1.0, 0.1) : vec4(0.5, 1.0, 1.0, 0.05);
    vec4 LF2COLOR = sunvvv ? vec4(0.42, 0.0, 1.0, 0.1): vec4(0.0, 0.42, 1.0, 0.05);
    vec4 LF3COLOR = sunvvv ? vec4(0.0, 1.0, 0.0, 0.1): vec4(0.0, 0.0, 1.0, 0.05);
    vec4 LF4COLOR = sunvvv ? vec4(1.0, 0.0, 0.0, 0.1): vec4(0.0, 0.5, 1.0, 0.05);
    
    LENS_FLARE(color, uv, lf1Pos, LF1SIZE, LF1COLOR);
    LENS_FLARE(color, uv, lf2Pos, LF2SIZE, LF2COLOR);
    LENS_FLARE(color, uv, lf3Pos, LF3SIZE, LF3COLOR);
    LENS_FLARE(color, uv, lf4Pos, LF4SIZE, LF4COLOR);
    return color;
}
#endif

vec3 randDir(in vec2 p)
{
	const float C = 65536.0 / 3.14159265;
	vec4 v = (1.0 / 65536.0) * p.xyxy + vec4(0.1, 0.2, 0.3, 0.4);

	v = fract(v.yzwx * C + v.xyzw);
	v = fract(v.yzwx * C + v.xyzw);
	v = fract(v.yzwx * C + v.xyzw);
	v = fract(v.yzwx * C + v.xyzw);

    v = v - mod(v, 1.0 / 15.0);
	return fract(v + frameCounter * 4.0 / 15.0);
}

void main()
{
    vec4 pack0 = texture(colortex0, Texcoord);
    vec4 pack1 = texture(colortex1, Texcoord);
    vec4 pack2 = texture(colortex2, Texcoord);
    vec4 pack6 = texture(colortex6, Texcoord);
    
    GbufferData gbuffer;
    ReadGbufferData(pack0, pack1, pack2, pack6, gbuffer);

    vec3 viewPos0 = getView(Texcoord, 0);
    vec3 viewPos1 = getView(Texcoord, 1);

    vec4 outColor = texture(colortex3, Texcoord);
    
    vec3 worldPos0 = View2World(viewPos0);
    float worldLen = sqrt(pow2(viewPos0.x) + pow2(viewPos0.y) + pow2(viewPos0.z));
    vec3 viewDir = viewPos0 / worldLen;
    vec3 worldDir = worldPos0 / worldLen;
    float dither = calcDither();
    if(gbuffer.materialID == 4)
    {
        vec2 warp = (vec2(gbuffer.ao, 1.0 - gbuffer.ao) * 2.0 - 1.0) * (1.0 - exp(-distance(viewPos0, viewPos1)));
        outColor.r = texture(colortex3, Texcoord + warp * 0.05).r;
        outColor.g = texture(colortex3, Texcoord + warp * 0.06).g;
        outColor.b = texture(colortex3, Texcoord + warp * 0.07).b;

        vec3 refDir = reflect(viewDir, gbuffer.Normal);
        vec3 refWorldDir = viewMat3Inverse * refDir;
        //float dither = fract(hash(gl_FragCoord.xy) + float(frameCounter) / 4.1672575);

        bool rr = true;
        
        #ifdef SSR
        vec3 refCoord = RayTracTexcoord(viewPos0 + gbuffer.Normal * (-viewPos0.z / far * 0.4 + 0.05), refDir, dither);
        rr = refCoord.z < 0.5;
        #endif
        vec3 refCol;

        if(rr)
        {
            if(isEyeInWater == 0)
            {
                refCol = DrawSky(refWorldDir, worldSunVec, true) * gbuffer.Imcoord.y;
            #ifdef VOLUME_BLOCK_CLOUDS
                if(refWorldDir.y > 0.0 && gbuffer.Imcoord.y > 0.0)
                {
                    float scatray = pow16(max(dot(shadowVec, refDir), 0.0) * 0.7 + 0.3) * min(dot(shadowVec, upVec) * 16.0, 1.0);
                    vec2 refClouds = volumeCloudRender(refWorldDir, worldLightVec, dither * 2.0 - 1.0, 9, 3);
                    refCol = mix(refCol, skyCloudAmbient + skyCloudLight * (0.2 + scatray * 1.35) * refClouds.y, refClouds.x * gbuffer.Imcoord.y);
                }
            #endif
            }else{
                refCol = waterFogColor;
            }
        }else{
        #ifdef SSR
            refCol = texture(colortex3, refCoord.xy).rgb;
        #endif
        }

        float cosTheta = dot(-viewDir, gbuffer.Normal);
        float fakeFresnel = pow4(1.0 - cosTheta);
        fakeFresnel = isEyeInWater == 0 ? fakeFresnel * 0.96 + 0.04 : smoothstep(0.01, 0.175, fakeFresnel);
        outColor.rgb = mix(outColor.rgb, refCol, fakeFresnel);

        
        vec3 shadowPos = World2ShadowProj(worldPos0);
        
        float NdotL = dot(gbuffer.Normal, shadowVec);
        float sdist = length(shadowPos.xy * 2.0 - 1.0);
        float offset = 8e-4 / (1.0 - sdist) * sqrt(1.0 - NdotL * NdotL);
        shadowPos.z -= offset;

        float highlightmask = textureLod(shadowtex1, shadowPos, 1.0) * min(dot(shadowVec, upVec) * 10.0, 1.0);

        outColor.rgb += skySunColor * min(DistributionGGX(gbuffer.Normal, normalize(shadowVec - viewDir), 0.125), 6.24) * highlightmask;

    }
    else if(gbuffer.smoothness > 0.0){
        vec3 refNorm = normalize(gbuffer.Normal + randDir(gl_FragCoord.xy) * (1.0 - gbuffer.smoothness) * 0.5);
        //vec3 refNorm = normalize(gbuffer.Normal + vec3(0.0, 0.0, (1.0 - gbuffer.smoothness) * 0.5 * dither));
        vec3 refDir = reflect(viewDir, refNorm);
        vec3 refWorldDir = viewMat3Inverse * refDir;
        
        bool rr = true;
        
        #ifdef SSR
        vec3 refCoord = RayTracTexcoord(viewPos0 + gbuffer.Normal * (-viewPos0.z / far * 0.4 + 0.05), refDir, dither);
        rr = refCoord.z < 0.5;
        #endif
        vec3 refCol;
        
        if(rr)
        {
            if(isEyeInWater == 0)
            {
                refCol = DrawSky(refWorldDir, worldSunVec, true) * gbuffer.Imcoord.y;
            #ifdef VOLUME_BLOCK_CLOUDS
                if(refWorldDir.y > 0.0 && gbuffer.Imcoord.y > 0.0)
                {
                    float scatray = pow16(max(dot(shadowVec, refDir), 0.0) * 0.7 + 0.3) * min(dot(shadowVec, upVec) * 16.0, 1.0);
                    vec2 refClouds = volumeCloudRender(refWorldDir, worldLightVec, dither * 2.0 - 1.0, 9, 3);
                    refCol = mix(refCol, skyCloudAmbient + skyCloudLight * (0.2 + scatray * 1.35) * refClouds.y, refClouds.x * gbuffer.Imcoord.y);
                }
            #endif
            }else{
                refCol = waterFogColor;
            }
        }else{
        #ifdef SSR
        for(int i = -2;i < 3;i++)
        {
            for(int j = -2;j < 3;j++)
            {
                refCol += texture(colortex3, refCoord.xy + vec2(i, j) * screenTexel * 4.0).rgb;
            }
        }
            refCol /= 25;
        #endif
        }
        outColor.rgb = BRDF_Reflect(gbuffer.Normal, -viewDir, shadowVec, outColor.rgb, gbuffer.Albedo.rgb, refCol, gbuffer.metalness, gbuffer.smoothness);

    }

    if(gbuffer.maskID == 1 && isEyeInWater == 0)
    {
        vec3 fogCol = DrawSky(worldDir, worldSunVec, false);
        float eyeB = smoothstep(0.0, 120.0, float(eyeBrightnessSmooth.y));
        outColor.rgb = mix(fogCol, outColor.rgb, exp(worldLen * (-0.008671875 * rainStrength - 0.001953125)) * eyeB + (1.0 - eyeB));
    }

    vec4 rainsnow = texture(colortex4, Texcoord);
    outColor.rgb = mix(outColor.rgb, vec3(rainsnow.r), rainsnow.g);

    #ifdef VOLUME_RAY
    float volumeRay = VolumeRayRender(worldDir, gbuffer.maskID == 1 ? worldLen : 1024.0, dither, gbuffer.maskID == 2);
    outColor.rgb += skySunColor * (isEyeInWater == 1 ? normalize(waterFogColor) : vec3(1.0)) * volumeRay * volumeRayStrength * VolumeRayTotalStrength;
    #endif

    #ifdef LENSFLARE
    outColor.rgb = lensFlare(outColor.rgb, Texcoord);
    #endif

    texBuffer3 = outColor;
}
/* DRAWBUFFERS:3 */