#version 460 compatibility

layout(location = 0) out vec4 texBuffer0;
layout(location = 1) out vec4 texBuffer1;
layout(location = 2) out vec4 texBuffer2;
layout(location = 3) out vec4 texBuffer6;


uniform sampler2D tex;

#ifdef MC_NORMAL_MAP
    uniform sampler2D normals;
#endif

#ifdef MC_SPECULAR_MAP
    uniform sampler2D specular;
#endif

#if MC_VERSION >= 11700
    uniform float alphaTestRef;
#endif

flat in float matID;
in float ao;
in vec2 Texcoord;
in vec2 Imcoord;
flat in vec2 texRes;
in vec3 vColor;
in vec3 TexelPos;

flat in vec3 vTangent;
flat in vec3 vBitangent;
flat in vec3 vNormal;

#ifndef MC_SPECULAR_MAP
    flat in vec4 specular_default;
#endif


#include "/HmwsLib/PackDatas.glsl"

#define TERRIAN_PARALLAX_ENABLE
#define TERRIAN_PARALLAX_MAXDEPTH 0.3//[0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0]

#include "/HmwsLib/TerrianParallax.glsl"

#define AUTO_NORMAL_SPAWN
#define HARDCODED_SPECULAR_DATA

float HeightMapping(vec3 x)
{
    return 1.0 - dot(x, vec3(0.2126, 0.7152, 0.0722));
}

const float TEXEL_AA_LOD_CONSERVATIVE = -1.0;
const float TEXEL_AA_LOD_RELAXED = 2.0;

vec4 bilinearSampler(in sampler2D source, in vec2 originalUV)
{
    originalUV -= 0.5;
    vec2 fractUv = fract(originalUV * atlasSize);
    fractUv = fractUv * fractUv * fractUv * (fractUv * (fractUv * 6.0 - 15.0) + 10.0);
    vec4 floorUv = vec4(floor(originalUV * atlasSize) / atlasSize, ceil(originalUV * atlasSize) / atlasSize);
    vec4 sample1 = texture(source, clampOffset(floorUv.xy));
    vec4 sample2 = texture(source, clampOffset(floorUv.zy));
    vec4 sample3 = texture(source, clampOffset(floorUv.xw));
    vec4 sample4 = texture(source, clampOffset(floorUv.zw));
    return mix(mix(sample1, sample2, fractUv.x), mix(sample3, sample4, fractUv.x), fractUv.y);    
}

vec4 texture_AA(in sampler2D source, in vec2 originalUV)
{
	vec2 dUV_dX = dFdx(originalUV) * atlasSize;
	vec2 dUV_dY = dFdy(originalUV) * atlasSize;

	vec2 delU = vec2(dUV_dX.x, dUV_dY.x);
	vec2 delV = vec2(dUV_dX.y, dUV_dY.y);
	vec2 adjustmentScalar = max(1.0 / vec2(length(delU), length(delV)), 1.0);

	vec2 fractionalTexel = fract(originalUV * atlasSize);
	vec2 adjustedFractionalTexel = clamp(fractionalTexel * adjustmentScalar, 0.0, 0.5) + clamp(fractionalTexel * adjustmentScalar - (adjustmentScalar - 0.5), 0.0, 0.5);

	float lod = log2(sqrt(max(dot(dUV_dX, dUV_dX), dot(dUV_dY, dUV_dY))) * 2.0);
	float samplingMode = smoothstep(TEXEL_AA_LOD_RELAXED, TEXEL_AA_LOD_CONSERVATIVE, lod);

	vec2 adjustedUV = (adjustedFractionalTexel + floor(originalUV * atlasSize)) / atlasSize;
	vec4 blendedSample = bilinearSampler(source, mix(originalUV, adjustedUV, samplingMode));

	return blendedSample;
}

vec3 GetNormalFromTex(vec2 texcoord)
{
    const float w = 1.0;
    float h0 = HeightMapping(texture(tex, texcoord).xyz);
    float h1 = HeightMapping(texture(tex, clampOffset(texcoord + vec2(w / atlasSize.x, 0.0))).xyz) - h0;
    float h2 = HeightMapping(texture(tex, clampOffset(texcoord + vec2(0.0, w / atlasSize.y))).xyz) - h0;
    float delta = abs(h1 + h2);
    delta = 1.0 - delta * delta;
    return normalize(vec3(h1, h2, delta));
}

void main()
{
    GbufferData buff;

    #if defined(MC_NORMAL_MAP) && defined(TERRIAN_PARALLAX_ENABLE)
    #ifdef TERRIAN_PARALLAX_ENABLE
        vec3 parallaxCoord = ParallaxOffsetCoord(Texcoord, TexelPos);
        vec2 instanceCoord = clampOffset(parallaxCoord.xy);
    #endif
    #else
    vec2 instanceCoord = Texcoord;
    #endif
    buff.Albedo = texture(tex, instanceCoord);
    buff.Albedo.rgb *= vColor;

    #if MC_VERSION >= 11700
        if(buff.Albedo.a < alphaTestRef)
            discard;
    #endif
    //buff.Albedo.rgb = vec3(parallaxCoord.z > 0.2 ? 1.0 : 0.0);

    buff.maskID = 1;
    buff.materialID = int(matID + 0.001);
    buff.ao = pow(ao, 2.2);
    
    #ifdef MC_NORMAL_MAP
    buff.Normal = mat3(vTangent, vBitangent, vNormal) * normalize(texture(normals, instanceCoord).xyz * 2.0 - 1.0);
    #else
    #ifdef AUTO_NORMAL_SPAWN
    buff.Normal = mat3(vTangent, vBitangent, vNormal) * GetNormalFromTex(Texcoord);
    #else
    buff.Normal = Normal;
    #endif
    #endif

    #ifdef MC_SPECULAR_MAP
    buff.Specular = texture(specular, instanceCoord);
    #else
    #ifdef HARDCODED_SPECULAR_DATA
    buff.Specular = specular_default;
    #else
    buff.Specular = vec4(0.0);
    #endif
    #endif

    buff.Imcoord = Imcoord;
    
    PackGbufferData(texBuffer0, texBuffer1, texBuffer2, texBuffer6, buff);
}
/* DRAWBUFFERS:0126 */